'''Contstants for NSQ'''

# NSQ Magic
MAGIC_V2 = '  V2'

# The newline character
NL = '\n'

# Response
FRAME_TYPE_RESPONSE = 0
FRAME_TYPE_ERROR = 1
FRAME_TYPE_MESSAGE = 2

# Command names
IDENTIFY = 'IDENTIFY'
SUB = 'SUB'
PUB = 'PUB'
MPUB = 'MPUB'
RDY = 'RDY'
FIN = 'FIN'
REQ = 'REQ'
TOUCH = 'TOUCH'
CLS = 'CLS'
NOP = 'NOP'

# Heartbeat text
HEARTBEAT = '_heartbeat_'



'''A class for interacting with a nsqlookupd instance over http'''

from . import BaseClient, json_wrap, ok_check


class Client(BaseClient):
    '''A client for talking to nsqlookupd over http'''
    @ok_check
    def ping(self):
        '''Ping the client'''
        return self.get('/ping')

    @json_wrap
    def info(self):
        '''Get info about this instance'''
        return self.get('/info')

    @json_wrap
    def lookup(self, topic):
        '''Look up which hosts serve a particular topic'''
        return self.get('/lookup', params={'topic': topic})

    @json_wrap
    def topics(self):
        '''Get a list of topics'''
        return self.get('/topics')

    @json_wrap
    def channels(self, topic):
        '''Get a list of channels for a given topic'''
        return self.get('/channels', params={'topic': topic})

    @json_wrap
    def nodes(self):
        '''Get information about nodes'''
        return self.get('/nodes')

    @json_wrap
    def delete_topic(self, topic):
        '''Delete a topic'''
        return self.get('/delete_topic', params={'topic': topic})

    @json_wrap
    def delete_channel(self, topic, channel):
        '''Delete a channel in the provided topic'''
        return self.get('/delete_channel',
            params={'topic': topic, 'channel': channel})

    @json_wrap
    def tombstone_topic_producer(self, topic, node):
        '''It's not clear what this endpoint does'''
        return self.get('/tombstone_topic_producer',
            params={'topic': topic, 'node': node})

    @json_wrap
    def create_topic(self, topic):
        '''Create a topic'''
        return self.get('/create_topic', params={'topic': topic})

    @json_wrap
    def create_channel(self, topic, channel):
        '''Create a channel in the provided topic'''
        return self.get('/create_channel',
            params={'topic': topic, 'channel': channel})

    @json_wrap
    def debug(self):
        '''Get debugging information'''
        return self.get('/debug')



#! /usr/bin/env python

import mock

import errno
import socket
import struct

from nsq import connection
from nsq import constants
from nsq import response
from nsq import util
from nsq import sockets
from common import FakeServerTest, IntegrationTest


class TestConnection(FakeServerTest):
    '''Test our connection class'''
    def connect(self):
        '''Return a connection / client'''
        return connection.Connection('localhost', self.ports[0], 0.01)

    def pack(self, frame, message):
        '''Pack a message'''
        return struct.pack('>ll', len(message) + 4, frame) + message

    def read(self, length):
        '''Read the provided length from the server'''
        found = ''
        while len(found) < length:
            try:
                found += self.servers[0].read(length)
            except socket.timeout:
                return found
        return found

    def send(self, message):
        '''Send a message from the server'''
        self.servers[0].send(message)

    def test_alive(self):
        '''Alive should return True if connected'''
        with self.identify():
            self.assertTrue(self.client.alive())

    def test_close(self):
        '''Should mark the connection as closed'''
        with self.identify():
            self.client.close()
            self.assertFalse(self.client.alive())

    def test_blocking(self):
        '''Sets blocking on the socket'''
        with self.identify():
            with mock.patch.object(self.client, '_socket') as mock_socket:
                self.client.setblocking(0)
                mock_socket.setblocking.assert_called_with(0)

    def test_pending(self):
        '''Appends to pending'''
        with self.identify():
            with mock.patch.object(self.client, '_socket') as mock_socket:
                mock_socket.send.return_value = 0
                self.client.setblocking(0)
                self.client.nop()
                self.assertEqual(self.client.pending(),
                    [constants.NOP + constants.NL])

    def test_flush_partial(self):
        '''Keeps its place when flushing out partial messages'''
        # We'll tell the connection it has only sent one byte when flushing
        with self.identify():
            with mock.patch.object(self.client._socket, 'send'):
                self.client._socket.send.return_value = 1
                self.client.setblocking(0)
                # Ensure this doesn't eagerly invoke our normal flush
                with mock.patch.object(self.client, 'flush'):
                    self.client.nop()
                self.client.flush()
                # We expect all but the first byte to remain
                message = constants.NOP + constants.NL
                self.assertEqual(self.client.pending(), [message[1:]])

    def test_flush_full(self):
        '''Pops off messages it has flushed completely'''
        with self.identify():
            message = constants.NOP + constants.NL
            # We'll tell the connection it has only sent one byte when flushing
            with mock.patch.object(self.client._socket, 'send', mock.Mock()):
                self.client._socket.send.return_value = len(message)
                self.client.setblocking(0)
                self.client.nop()
                self.client.flush()
                # The nop message was sent, so we expect it to be popped
                self.assertEqual(self.client.pending(), [])

    def test_flush_count(self):
        '''Returns how many bytes were sent'''
        with self.identify():
            message = constants.NOP + constants.NL
            with mock.patch.object(self.client._socket, 'send', mock.Mock()):
                self.client._socket.send.return_value = len(message)
                self.client.setblocking(0)
                # Ensure this doesn't invoke our normal flush
                with mock.patch.object(self.client, 'flush'):
                    self.client.nop()
                self.assertEqual(self.client.flush(), len(message))

    def test_flush_empty(self):
        '''Returns 0 if there are no pending messages'''
        with self.identify():
            self.client.setblocking(0)
            self.assertEqual(self.client.flush(), 0)

    def test_flush_multiple(self):
        '''Flushes as many messages as possible'''
        with self.identify():
            with mock.patch.object(self.client, '_pending', ['hello'] * 5):
                with mock.patch.object(
                    self.client._socket, 'send', return_value=5):
                    self.client.flush()
                self.assertEqual(len(self.client.pending()), 0)

    def test_flush_would_block(self):
        '''Honors EAGAIN / EWOULDBLOCK'''
        with self.identify():
            with mock.patch.object(self.client, '_socket') as mock_socket:
                with mock.patch.object(self.client, '_pending', [1, 2, 3]):
                    mock_socket.send.side_effect = socket.error(errno.EAGAIN)
                    self.assertEqual(self.client.flush(), 0)

    def test_flush_socket_error(self):
        '''Re-raises socket non-EAGAIN errors'''
        with self.identify():
            with mock.patch.object(self.client, '_socket') as mock_socket:
                with mock.patch.object(self.client, '_pending', [1, 2, 3]):
                    mock_socket.send.side_effect = socket.error('foo')
                    self.assertRaises(socket.error, self.client.flush)

    def test_eager_flush(self):
        '''Sending on a non-blocking connection eagerly flushes'''
        with self.identify():
            with mock.patch.object(self.client, 'flush') as mock_flush:
                self.client.setblocking(0)
                self.client.send('foo')
                mock_flush.assert_called_with()

    def test_magic(self):
        '''Sends the NSQ magic bytes'''
        # This does /not/ use identify, because we're interested in seeing that
        # the magic bytes are the first thing sent
        self.assertEqual(
            self.read(len(constants.MAGIC_V2)), constants.MAGIC_V2)

    def test_identify(self):
        '''The connection sends the identify commands'''
        # This does /not/ use identify, because we're interested in seeing that
        # the magic bytes are the first thing sent
        self.servers[0].assertMagic()
        identification = self.servers[0].readIdentify()
        self.assertEqual(identification, self.client._identify_options)

    def test_read_timeout(self):
        '''Returns no results after a socket timeout'''
        with self.identify():
            with mock.patch.object(self.client, '_socket') as mock_socket:
                mock_socket.recv.side_effect = socket.timeout
                self.assertEqual(self.client.read(), [])

    def test_read_socket_error(self):
        '''Re-raises socket non-errno socket errors'''
        with self.identify():
            with mock.patch.object(self.client, '_socket') as mock_socket:
                mock_socket.recv.side_effect = socket.error('foo')
                self.assertRaises(socket.error, self.client.read)

    def test_read_would_block(self):
        '''Returns no results if it would block'''
        with self.identify():
            with mock.patch.object(self.client, '_socket') as mock_socket:
                mock_socket.recv.side_effect = socket.error(errno.EAGAIN)
                self.client.setblocking(0)
                self.assertEqual(self.client.read(), [])

    def test_read_partial(self):
        '''Returns nothing if it has only read partial results'''
        with self.identify():
            self.send('f')
            self.assertEqual(self.client.read(), [])

    def test_read_size_partial(self):
        '''Returns one response size is complete, but content is partial'''
        with self.identify():
            self.send(
                self.pack(constants.FRAME_TYPE_RESPONSE, 'hello')[:-1])
            self.assertEqual(self.client.read(), [])

    def test_read_whole(self):
        '''Returns a single message if it has read a complete one'''
        with self.identify():
            self.send(self.pack(constants.FRAME_TYPE_RESPONSE, 'hello'))
            expected = response.Response(
                self.client, constants.FRAME_TYPE_RESPONSE, 'hello')
            self.assertEqual(self.client.read(), [expected])

    def test_read_multiple(self):
        '''Returns multiple responses if available'''
        with self.identify():
            packed = self.pack(constants.FRAME_TYPE_RESPONSE, 'hello') * 10
            self.send(packed)
            expected = response.Response(
                self.client, constants.FRAME_TYPE_RESPONSE, 'hello')
            self.assertEqual(self.client.read(), [expected] * 10)

    def test_fileno(self):
        '''Returns the connection's file descriptor appropriately'''
        with self.identify():
            self.assertEqual(
                self.client.fileno(), self.client._socket.fileno())

    def test_str_alive(self):
        '''Sane str representation for an alive connection'''
        with self.identify():
            with mock.patch.object(self.client, 'alive', return_value=True):
                with mock.patch.object(
                    self.client, 'fileno', return_value=7):
                    with mock.patch.object(self.client, 'host', 'host'):
                        with mock.patch.object(self.client, 'port', 'port'):
                            self.assertEqual(str(self.client),
                                '<Connection host:port (alive on FD 7)>')

    def test_str_dead(self):
        '''Sane str representation for an alive connection'''
        with self.identify():
            with mock.patch.object(self.client, 'alive', return_value=False):
                with mock.patch.object(
                    self.client, 'fileno', return_value=7):
                    with mock.patch.object(self.client, 'host', 'host'):
                        with mock.patch.object(self.client, 'port', 'port'):
                            self.assertEqual(str(self.client),
                                '<Connection host:port (dead on FD 7)>')

    def test_send_no_message(self):
        '''Appropriately sends packed data without message'''
        with self.identify():
            self.client.nop()
            expected = constants.NOP + constants.NL
            self.assertEqual(self.read(len(expected)), expected)

    def test_send_message(self):
        '''Appropriately sends packed data with message'''
        with self.identify():
            self.client.identify({})
            expected = ''.join(
                (constants.IDENTIFY, constants.NL, util.pack('{}')))
            self.assertEqual(self.read(len(expected)), expected)

    def test_sub(self):
        '''Appropriately sends sub'''
        with self.identify():
            self.client.sub('foo', 'bar')
            expected = ''.join((constants.SUB, ' foo bar', constants.NL))
            self.assertEqual(self.read(len(expected)), expected)

    def test_pub(self):
        '''Appropriately sends pub'''
        with self.identify():
            self.client.pub('foo', 'hello')
            expected = ''.join(
                (constants.PUB, ' foo', constants.NL, util.pack('hello')))
            self.assertEqual(self.read(len(expected)), expected)

    def test_mpub(self):
        '''Appropriately sends mpub'''
        with self.identify():
            self.client.mpub('foo', 'hello', 'howdy')
            expected = ''.join((
                constants.MPUB, ' foo', constants.NL,
                util.pack(['hello', 'howdy'])))
            self.assertEqual(self.read(len(expected)), expected)

    def test_ready(self):
        '''Appropriately sends ready'''
        with self.identify():
            self.client.rdy(5)
            expected = ''.join((constants.RDY, ' 5', constants.NL))
            self.assertEqual(self.read(len(expected)), expected)

    def test_fin(self):
        '''Appropriately sends fin'''
        with self.identify():
            self.client.fin('message_id')
            expected = ''.join((constants.FIN, ' message_id', constants.NL))
            self.assertEqual(self.read(len(expected)), expected)

    def test_req(self):
        '''Appropriately sends req'''
        with self.identify():
            self.client.req('message_id', 10)
            expected = ''.join((constants.REQ, ' message_id 10', constants.NL))
            self.assertEqual(self.read(len(expected)), expected)

    def test_touch(self):
        '''Appropriately sends touch'''
        with self.identify():
            self.client.touch('message_id')
            expected = ''.join((constants.TOUCH, ' message_id', constants.NL))
            self.assertEqual(self.read(len(expected)), expected)

    def test_cls(self):
        '''Appropriately sends cls'''
        with self.identify():
            self.client.cls()
            expected = ''.join((constants.CLS, constants.NL))
            self.assertEqual(self.read(len(expected)), expected)

    def test_nop(self):
        '''Appropriately sends nop'''
        with self.identify():
            self.client.nop()
            expected = ''.join((constants.NOP, constants.NL))
            self.assertEqual(self.read(len(expected)), expected)

    # Some tests very closely aimed at identification
    def test_identify_response(self):
        '''We can get our identify response'''
        expected = {'foo': 'bar'}
        with self.identify(expected) as responses:
            self.assertEqual(len(responses), 1)
            self.assertEqual(responses[0].data, expected)

    def test_calls_identified(self):
        '''Upon getting an identification response, we call 'identified'''
        with mock.patch.object(self.client, 'identified') as mock_identified:
            with self.identify():
                self.assertTrue(mock_identified.called)

    def test_identified_sets_received(self):
        '''Upon getting an identification response, we set identify_received'''
        self.assertFalse(self.client._identify_received)
        with self.identify():
            self.assertTrue(self.client._identify_received)

    def test_identified_tolerates_ok(self):
        '''The identified handler tolerates OK responses'''
        res = mock.Mock(data='OK')
        self.assertEqual(self.client.identified(res).data, 'OK')

    def test_identify_defaults(self):
        '''Identify provides default options'''
        self.assertEqual(self.client._identify_options, {
            'feature_negotiation': True,
            'long_id': socket.getfqdn(),
            'short_id': socket.gethostname(),
            'user_agent': self.client.USER_AGENT
        })

    def test_identify_override_defaults(self):
        '''Identify allows us to override defaults'''
        with mock.patch('nsq.connection.Connection.connect'):
            conn = connection.Connection('host', 0, long_id='not-your-fqdn')
            self.assertEqual(conn._identify_options['long_id'], 'not-your-fqdn')

    def test_identify_tls_unsupported(self):
        '''Raises an exception about the lack of TLS support'''
        with mock.patch('nsq.connection.TLSSocket', None):
            self.assertRaises(
                AssertionError, connection.Connection, 'host', 0, tls_v1=True)

    def test_identify_snappy_unsupported(self):
        '''Raises an exception about the lack of snappy support'''
        with mock.patch('nsq.connection.SnappySocket', None):
            self.assertRaises(
                AssertionError, connection.Connection, 'host', 0, snappy=True)

    def test_identify_deflate_unsupported(self):
        '''Raises an exception about the lack of deflate support'''
        with mock.patch('nsq.connection.DeflateSocket', None):
            self.assertRaises(
                AssertionError, connection.Connection, 'host', 0, deflate=True)

    def test_identify_no_deflate_level(self):
        '''Raises an exception about the lack of deflate_level support'''
        with mock.patch('nsq.connection.DeflateSocket', None):
            self.assertRaises(AssertionError,
                connection.Connection, 'host', 0, deflate_level=True)

    def test_identify_no_snappy_and_deflate(self):
        '''We should yell early about incompatible snappy and deflate options'''
        self.assertRaises(AssertionError,
            connection.Connection, 'host', 0, snappy=True, deflate=True)

    def test_identify_saves_identify_response(self):
        '''Saves the identify response from the server'''
        expected = {'foo': 'bar'}
        with self.identify(expected):
            self.assertEqual(self.client._identify_response, expected)

    def test_identify_saves_max_rdy_count(self):
        '''Saves the max ready count if it's provided'''
        with self.identify({'max_rdy_count': 100}):
            self.assertEqual(self.client.max_rdy_count, 100)



from . import backoff
from . import constants
from . import logger
from . import response
from . import util
from . import json
from . import __version__
from .sockets import TLSSocket, SnappySocket, DeflateSocket

import errno
import socket
import struct
import sys


class Connection(object):
    '''A socket-based connection to a NSQ server'''
    # Default user agent
    USER_AGENT = 'nsq-py/%s' % __version__

    def __init__(self, host, port, timeout=1.0, reconnection_backoff=None,
        **identify):
        assert isinstance(host, (str, unicode))
        assert isinstance(port, int)
        self._socket = None
        self._buffer = ''
        # Our host and port
        self.host = host
        self.port = port
        # Whether or not our socket is set to block
        self._blocking = 1
        # The pending messages we have to send
        self._pending = []
        self._timeout = timeout
        # The last ready time we set our ready count, current ready count
        self.last_ready_sent = 0
        self.ready = 0
        # Whether or not we've received an identify response
        self._identify_received = False
        self._identify_response = {}
        # The options to use when identifying
        self._identify_options = dict(identify)
        self._identify_options.setdefault('short_id', socket.gethostname())
        self._identify_options.setdefault('long_id', socket.getfqdn())
        self._identify_options.setdefault('feature_negotiation', True)
        self._identify_options.setdefault('user_agent', self.USER_AGENT)

        # Some settings that may be determined by an identify response
        self.max_rdy_count = sys.maxint

        # Check for any options we don't support
        disallowed = []
        if not SnappySocket:  # pragma: no branch
            disallowed.append('snappy')
        if not DeflateSocket:  # pragma: no branch
            disallowed.extend(['deflate', 'deflate_level'])
        if not TLSSocket:  # pragma: no branch
            disallowed.append('tls_v1')
        for key in disallowed:
            assert not self._identify_options.get(key, False), (
                'Option %s is not supported' % key)

        # Our backoff policy for reconnection. The default is to use an
        # exponential backoff 8 * (2 ** attempt) clamped to [0, 60]
        self._reconnection_backoff = (
            reconnection_backoff or
            backoff.Clamped(backoff.Exponential(2, 8), maximum=60))
        self._reconnnection_counter = backoff.ResettingAttemptCounter(
            self._reconnection_backoff)

        # Establish our connection
        self.connect()

    def __str__(self):
        state = 'alive' if self.alive() else 'dead'
        return '<Connection %s:%s (%s on FD %s)>' % (
            self.host, self.port, state, self.fileno())

    def ready_to_reconnect(self):
        '''Returns True if enough time has passed to attempt a reconnection'''
        return self._reconnnection_counter.ready()

    def connect(self):
        '''Establish a connection'''
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.settimeout(self._timeout)
        self._socket.connect((self.host, self.port))
        # Set our socket's blocking state to whatever ours is
        self._socket.setblocking(self._blocking)
        # Safely write our magic
        self._pending.append(constants.MAGIC_V2)
        self.flush()
        # And send our identify command
        self.identify(self._identify_options)
        # At this point, we've not received an identify response
        self._identify_received = False

    def close(self):
        '''Close our connection'''
        try:
            if self._socket:
                self._socket.close()
        finally:
            self._socket = None

    def read(self):
        '''Read from the socket, and return a list of responses'''
        # It's important to know that it may return no responses or multiple
        # responses. It depends on how the buffering works out. First, read from
        # the socket
        try:
            packet = self._socket.recv(4096)
        except socket.timeout:
            # If the socket times out, return nothing
            return []
        except socket.error as exc:
            # Catch (errno, message)-type socket.errors
            if exc.args[0] == errno.EAGAIN:
                return []
            else:
                raise

        # Append our newly-read data to our buffer
        logger.debug('Read %s from socket', packet)
        self._buffer += packet

        responses = []
        while len(self._buffer) >= 4:
            size = struct.unpack('>l', self._buffer[:4])[0]
            logger.debug('Read size of %s', size)
            # Now check to see if there's enough left in the buffer to read the
            # message.
            if (len(self._buffer) - 4) >= size:
                message = self._buffer[4:(size + 4)]
                res = response.Response.from_raw(self, message)
                if isinstance(res, response.Message):
                    self.ready -= 1
                elif not self._identify_received:
                    # Handle the identify response if we've not yet received it
                    if isinstance(res, response.Response):  # pragma: no branch
                        res = self.identified(res)
                responses.append(res)
                self._buffer = self._buffer[(size + 4):]
            else:
                break
        return responses

    def identified(self, res):
        '''Handle a response to our 'identify' command. Returns response'''
        # If they support it, they should give us a JSON blob which we should
        # inspect.
        try:
            res.data = json.loads(res.data)
            self._identify_response = res.data
            # Save our max ready count unless it's not provided
            self.max_rdy_count = res.data.get(
                'max_rdy_count', self.max_rdy_count)
        except:
            pass
        finally:
            # Always mark that we've now handled the receive
            self._identify_received = True
        return res

    def alive(self):
        '''Returns True if this connection is alive'''
        return bool(self._socket)

    def setblocking(self, blocking):
        '''Set whether or not this message is blocking'''
        self._socket.setblocking(blocking)
        self._blocking = blocking

    def fileno(self):
        '''Returns the socket's fileno. This allows us to select on this'''
        return self._socket.fileno()

    def pending(self):
        '''All of the messages waiting to be sent'''
        return self._pending

    def flush(self):
        '''Flush some of the waiting messages, returns count written'''
        # We can only send at most one message here, because all we know is
        # that the socket can have some data written to it. We don't know how
        # many messages-worth might be sent. An alternative would be to keep
        # around a single string of the data that remains to be sent so that we
        # could potentially send larger messages
        total = 0
        while self._pending:
            try:
                # Try to send as much of the first message as possible
                count = self._socket.send(self._pending[0])
                if count < len(self._pending[0]):
                    # Save the rest of the message that could not be sent
                    self._pending[0] = self._pending[0][count:]
                    break
                else:
                    # Otherwise, pop off this message
                    self._pending.pop(0)
                    total += count
            except socket.error as exc:
                # Catch (errno, message)-type socket.errors
                if exc.args[0] == errno.EAGAIN:
                    break
                else:
                    raise
        return total

    def send(self, command, message=None):
        '''Send a command over the socket with length endcoded'''
        if message:
            joined = command + constants.NL + util.pack(message)
        else:
            joined = command + constants.NL
        if self._blocking:
            self._socket.sendall(joined)
        else:
            self._pending.append(joined)
            self.flush()

    def identify(self, data):
        '''Send an identification message'''
        return self.send(constants.IDENTIFY, json.dumps(data))

    def sub(self, topic, channel):
        '''Subscribe to a topic/channel'''
        return self.send(' '.join((constants.SUB, topic, channel)))

    def pub(self, topic, message):
        '''Publish to a topic'''
        return self.send(' '.join((constants.PUB, topic)), message)

    def mpub(self, topic, *messages):
        '''Publish multiple messages to a topic'''
        return self.send(constants.MPUB + ' ' + topic, messages)

    def rdy(self, count):
        '''Indicate that you're ready to receive'''
        self.ready = count
        self.last_ready_sent = count
        return self.send(constants.RDY + ' ' + str(count))

    def fin(self, message_id):
        '''Indicate that you've finished a message ID'''
        return self.send(constants.FIN + ' ' + message_id)

    def req(self, message_id, timeout):
        '''Re-queue a message'''
        return self.send(constants.REQ + ' ' + message_id + ' ' + str(timeout))

    def touch(self, message_id):
        '''Reset the timeout for an in-flight message'''
        return self.send(constants.TOUCH + ' ' + message_id)

    def cls(self):
        '''Close the connection cleanly'''
        return self.send(constants.CLS)

    def nop(self):
        '''Send a no-op'''
        return self.send(constants.NOP)



'''Test all of our utility functions'''

import unittest

import struct
from nsq import util


class TestPack(unittest.TestCase):
    '''Test our packing utility'''
    def test_string(self):
        '''Give it a low-ball test'''
        message = 'hello'
        self.assertEqual(util.pack(message), struct.pack('>l5s', 5, message))

    def test_iterable(self):
        '''Make sure it handles iterables'''
        messages = ['hello'] * 10
        packed = struct.pack('>l5s', 5, 'hello')
        expected = struct.pack('>ll90s', 94, 10, packed * 10)
        self.assertEqual(util.pack(messages), expected)


class TestHexify(unittest.TestCase):
    '''Test our hexification utility'''
    def setUp(self):
        self.message = '\x00hello\n\tFOO2'

    def test_identical(self):
        '''Does not transform the value of the text'''
        import ast
        hexified = util.hexify(self.message)
        self.assertEqual(self.message, ast.literal_eval("'%s'" % hexified))

    def test_meaningful(self):
        '''The output it gives is meaningful'''
        hexified = util.hexify(self.message)
        self.assertEqual(hexified, '\\x00hello\\x0a\\x09FOO2')


class TestDistribute(unittest.TestCase):
    '''Test the distribute'''
    def counts(self, total, objects):
        '''Return a list of the counts returned by distribute'''
        return zip(*util.distribute(total, objects))[0]

    def count(self, total, objects):
        '''Return the sum of the counts'''
        return sum(self.counts(total, objects))

    def test_sum_evenly_divisible(self):
        '''We get the expected total when total is evenly divisible'''
        self.assertEqual(self.count(10, range(5)), 10)

    def test_sum_not_evenly_divisible(self):
        '''We get the expected total when total not evenly divisible'''
        self.assertEqual(self.count(10, range(3)), 10)

    def test_min_max(self):
        '''The minimum and maximum should be within 1'''
        for num in xrange(1, 50):
            objects = range(num)
            for total in xrange(1, 50):
                counts = self.counts(total, objects)
                self.assertLessEqual(max(counts) - min(counts), 1)



'''Our clients for interacting with various clients'''

from decorator import decorator
import requests

from .. import json
from ..exceptions import NSQException


@decorator
def wrap(function, *args, **kwargs):
    '''Wrap a function that returns a request with some exception handling'''
    try:
        req = function(*args, **kwargs)
        if req.status_code == 200:
            return req
        else:
            raise ClientException(req.reason, req.content)
    except ClientException:
        raise
    except Exception as exc:
        raise ClientException(exc)


@decorator
def json_wrap(function, *args, **kwargs):
    '''Return the json content of a function that returns a request'''
    try:
        # Some responses have data = None, but they generally signal a
        # successful API call as well.
        return json.loads(function(*args, **kwargs).content)['data'] or True
    except Exception as exc:
        raise ClientException(exc)


@decorator
def ok_check(function, *args, **kwargs):
    '''Ensure that the response body is OK'''
    req = function(*args, **kwargs)
    if req.content.lower() != 'ok':
        raise ClientException(req.content)
    return req.content


class ClientException(NSQException):
    '''An exception class for all client errors'''


class BaseClient(object):
    '''Base client class'''
    def __init__(self, host):
        '''Host may be a 'host:port' string or a (host, port) tuple'''
        if isinstance(host, basestring):
            # Strip off the scheme if any was provideds
            _, __, hostname = host.partition('//')
            self._host, _, self._port = hostname.partition(':')
        elif isinstance(host, (tuple, list)):
            self._host, self._port = host
        else:
            raise TypeError('Host must be a string or tuple')
        assert self._host, 'Must provide a host'
        assert self._port, 'Must provide a port'

    @wrap
    def get(self, path, *args, **kwargs):
        '''GET the provided endpoint'''
        return requests.get(
            'http://%s:%s%s' % (self._host, self._port, path), *args, **kwargs)

    @wrap
    def post(self, path, *args, **kwargs):
        '''POST to the provided endpoint'''
        return requests.post(
            'http://%s:%s%s' % (self._host, self._port, path), *args, **kwargs)



from .client import Client
from .response import Message
from .util import distribute
from . import logger


class Reader(Client):
    '''A client meant exclusively for reading'''
    def __init__(self, topic, channel, lookupd_http_addresses=None,
        nsqd_tcp_addresses=None, max_in_flight=200, **identify):
        self._channel = channel
        self._max_in_flight = max_in_flight
        Client.__init__(
            self, lookupd_http_addresses, nsqd_tcp_addresses, topic, **identify)

    def add(self, connection):
        '''Add this connection and manipulate its RDY state'''
        conn = Client.add(self, connection)
        if conn:
            conn.sub(self._topic, self._channel)
            conn.rdy(1)

    def distribute_ready(self):
        '''Distribute the ready state across all of the connections'''
        connections = [c for c in self.connections() if c.alive()]
        if len(connections) > self._max_in_flight:
            raise NotImplementedError(
                'Max in flight must be greater than number of connections')
        else:
            # Distribute the ready count evenly among the connections
            for count, conn in distribute(self._max_in_flight, connections):
                # We cannot exceed the maximum RDY count for a connection
                if count > conn.max_rdy_count:
                    logger.info(
                        'Using max_rdy_count (%i) instead of %i for %s RDY',
                        conn.max_rdy_count, count, conn)
                    count = conn.max_rdy_count
                logger.info('Sending RDY %i to %s', count, conn)
                conn.rdy(count)

    def needs_distribute_ready(self):
        '''Determine whether or not we need to redistribute the ready state'''
        # Try to pre-empty starvation by comparing current RDY against
        # the last value sent.
        alive = [c for c in self.connections()]
        if any(c.ready <= (c.last_ready_sent * 0.25) for c in alive):
            return True

    def close_connection(self, connection):
        '''A hook into when connections are closed'''
        Client.close_connection(self, connection)
        self.distribute_ready()

    def read(self):
        '''Read some number of messages'''
        found = Client.read(self)

        # Redistribute our ready state if necessary
        if self.needs_distribute_ready():
            self.distribute_ready()

        # Finally, return all the results we've read
        return found

    def __iter__(self):
        with self.connection_checker():
            while True:
                for message in self.read():
                    # A reader's only interested in actual messages
                    if isinstance(message, Message):
                        # We'll probably add a hook in here to track the RDY
                        # states of our connections
                        yield message



'''A class that checks connections'''

import time
import threading


class StoppableThread(threading.Thread):
    '''A thread that may be stopped'''
    def __init__(self):
        threading.Thread.__init__(self)
        self._event = threading.Event()

    def wait(self, timeout):
        '''Wait for the provided time to elapse'''
        return self._event.wait(timeout)

    def stop(self):
        '''Set the stop condition'''
        self._event.set()


class PeriodicThread(StoppableThread):
    '''A thread that periodically invokes a callback every interval seconds'''
    def __init__(self, interval, callback, *args, **kwargs):
        StoppableThread.__init__(self)
        self._interval = interval
        self._callback = callback
        self._args = args
        self._kwargs = kwargs
        self._last_checked = None

    def delay(self):
        '''How long to wait before the next check'''
        if self._last_checked:
            return self._interval - (time.time() - self._last_checked)
        return self._interval

    def callback(self):
        '''Run the callback'''
        self._callback(*self._args, **self._kwargs)
        self._last_checked = time.time()

    def run(self):
        '''Run the callback periodically'''
        while not self.wait(self.delay()):
            self.callback()


class ConnectionChecker(PeriodicThread):
    '''A thread that checks the connections on an object'''
    def __init__(self, client, interval=60):
        PeriodicThread.__init__(self, interval, client.check_connections)



import mock
import unittest

from nsq import client
from nsq import response
from nsq import constants
from nsq import exceptions
from nsq.http import ClientException

from common import FakeServerTest


class TestClientNsqd(FakeServerTest):
    '''Test our client class'''
    def connect(self):
        '''Return a new client'''
        hosts = ['localhost:%s' % port for port in self.ports]
        return client.Client(nsqd_tcp_addresses=hosts)

    def test_connect_nsqd(self):
        '''Can successfully establish connections'''
        with self.identify():
            connections = self.client.connections()
            self.assertEqual(len(connections), 1)
            for connection in connections:
                self.assertTrue(connection.alive())

    def test_add_existing(self):
        '''Adding an existing connection returns None'''
        with self.identify():
            connection = self.client.connections()[0]
            self.assertEqual(self.client.add(connection), None)

    def test_remove_exception(self):
        '''If closing a connection raises an exception, remove still works'''
        with self.identify():
            connection = self.client.connections()[0]
            with mock.patch.object(connection, 'close', side_effect=Exception):
                self.assertEqual(self.client.remove(connection), connection)

    def test_honors_identify_options(self):
        '''Sends along identify options to each connection as it's created'''
        with mock.patch('nsq.client.connection.Connection') as MockConnection:
            with mock.patch.object(
                self.client, '_identify_options', {'foo': 'bar'}):
                self.client.connect('foo', 'bar')
                MockConnection.assert_called_with('foo', 'bar',
                    reconnection_backoff=None, foo='bar')

    def test_conection_checker(self):
        '''Spawns and starts a connection checker'''
        with self.client.connection_checker() as checker:
            self.assertTrue(checker.is_alive())
        self.assertFalse(checker.is_alive())


class TestClientLookupd(FakeServerTest):
    '''Test our client class'''
    def connect(self):
        '''Return a new client'''
        with mock.patch('nsq.client.nsqlookupd.Client') as MockClass:
            MockClass.return_value.lookup.return_value = {
                'data': {
                    'producers': [{
                        'broadcast_address': 'localhost',
                        'tcp_port': self.ports[0]
                    }]
                }
            }
            return client.Client(topic='foo',
                lookupd_http_addresses=['http://localhost:1234'])

    def test_connected(self):
        '''Can successfully establish connections'''
        with self.identify():
            connections = self.client.connections()
            self.assertEqual(len(connections), 1)
            for connection in connections:
                self.assertTrue(connection.alive())

    def test_asserts_topic(self):
        '''If nsqlookupd servers are provided, asserts a topic'''
        self.assertRaises(
            AssertionError, client.Client, lookupd_http_addresses=['foo'])

    def test_client_exception(self):
        '''Is OK when discovery fails'''
        with mock.patch('nsq.client.nsqlookupd.Client') as MockClass:
            instance = MockClass.return_value
            instance.lookup.side_effect = ClientException
            self.client = client.Client(
                lookupd_http_addresses=['http://localhost:1234'],
                topic='foo')

    def test_discover_connected(self):
        '''Doesn't freak out when rediscovering established connections'''
        before = self.client.connections()
        self.client.discover('foo')
        self.assertEqual(self.client.connections(), before)

    def test_discover_closed(self):
        '''Reconnects to discovered servers that have closed connections'''
        for conn in self.client.connections():
            conn.close()
        state = [conn.alive() for conn in self.client.connections()]
        self.assertEqual(state, [False])
        self.client.discover('foo')
        state = [conn.alive() for conn in self.client.connections()]
        self.assertEqual(state, [True])


class TestClientMultiple(FakeServerTest):
    '''Tests for our client class'''
    ports = (12345, 12346)

    def connect(self):
        '''Return a new client'''
        hosts = ['localhost:%s' % port for port in self.ports]
        return client.Client(nsqd_tcp_addresses=hosts)

    def test_multi_read(self):
        '''Can read from multiple sockets'''
        with self.identify():
            self.servers[0].response('hello')
            self.servers[1].response('hello')
            found = self.client.read()
            self.assertEqual(len(found), 2)
            for res in found:
                self.assertIsInstance(res, response.Response)
                self.assertEqual(res.data, 'hello')

    def test_heartbeat(self):
        '''Sends a nop on connections that have received a heartbeat'''
        with self.identify():
            self.servers[0].response(constants.HEARTBEAT)
            # Get the heartbeat and automatically send the NOP
            self.assertEqual(self.client.read(), [])
            self.assertEqual(
                self.servers[0].read(100), constants.NOP + constants.NL)

    def test_closes_on_fatal(self):
        '''All but a few errors are considered fatal'''
        with self.identify():
            self.servers[0].error(exceptions.InvalidException)
            self.client.read()
            states = set(conn.alive() for conn in self.client.connections())
            self.assertEqual(states, set((True, False)))

    def test_nonfatal(self):
        '''Nonfatal errors keep the connection open'''
        with self.identify():
            self.servers[0].error(exceptions.FinFailedException)
            self.client.read()
            states = set(conn.alive() for conn in self.client.connections())
            self.assertEqual(states, set((True,)))

    def test_passes_errors(self):
        '''The client's read method should now swallow Error responses'''
        with self.identify():
            self.servers[0].error(exceptions.InvalidException)
            res = self.client.read()
            self.assertEqual(len(res), 1)
            self.assertIsInstance(res[0], response.Error)
            self.assertEqual(res[0].data, exceptions.InvalidException.name)

    def test_closes_on_exception(self):
        '''If a connection gets an exception, it closes it'''
        # Pick a connection to have throw an exception
        with self.identify():
            connection = self.client.connections()[0]
            with mock.patch.object(
                connection, 'read', side_effect=exceptions.NSQException):
                self.servers[0].response('hello')
                self.servers[1].response('hello')
                self.client.read()
                self.assertFalse(connection.alive())

    def test_read_writable(self):
        '''Read flushes any writable connections'''
        with self.identify():
            with mock.patch('nsq.client.select') as MockSelect:
                connection = mock.Mock()
                MockSelect.select.return_value = ([], [connection], [])
                self.client.read()
                connection.flush.assert_called_with()

    def test_read_exceptions(self):
        '''Read flushes any writable connections'''
        with self.identify():
            with mock.patch('nsq.client.select') as MockSelect:
                connection = mock.Mock()
                MockSelect.select.return_value = ([], [], [connection])
                self.client.read()
                connection.close.assert_called_with()

    def test_read_timeout(self):
        '''Logs a message when our read loop finds nothing because of timeout'''
        with self.identify():
            with mock.patch('nsq.client.select') as MockSelect:
                with mock.patch('nsq.client.logger') as MockLogger:
                    MockSelect.select.return_value = ([], [], [])
                    self.client.read()
                    MockLogger.debug.assert_called_with('Timed out...')

    def test_read_with_no_connections(self):
        '''Attempting to read with no connections'''
        with self.identify():
            with mock.patch.object(self.client, 'connections', return_value=[]):
                self.assertEqual(self.client.read(), [])

    def test_random_connection(self):
        '''Yields a random client'''
        found = []
        for _ in xrange(20):
            with self.client.random_connection() as conn:
                found.append(conn)
        self.assertEqual(set(found), set(self.client.connections()))

    def test_wait_response(self):
        '''Waits until a response is available'''
        with mock.patch.object(
            self.client, 'read', side_effect=[[], ['hello']]):
            self.assertEqual(self.client.wait_response(), ['hello'])

    def test_wait_write(self):
        '''Waits until a command has been sent'''
        connection = mock.Mock()
        with mock.patch.object(self.client, 'read'):
            connection.pending = mock.Mock(side_effect=[True, False])
            self.client.wait_write(connection)
            self.assertTrue(connection.pending.called)

    def test_pub(self):
        '''Pub called on a random connection and waits for a response'''
        connection = mock.Mock()
        with mock.patch.object(
            self.client, 'connections', return_value=[connection]):
            with mock.patch.object(
                self.client, 'wait_response', return_value=['response']):
                self.assertEqual(self.client.pub('foo', 'bar'), ['response'])
                connection.pub.assert_called_with('foo', 'bar')

    def test_mpub(self):
        '''Mpub called on a random connection and waits for a response'''
        connection = mock.Mock()
        messages = ['hello', 'how', 'are', 'you']
        with mock.patch.object(
            self.client, 'connections', return_value=[connection]):
            with mock.patch.object(
                self.client, 'wait_response', return_value=['response']):
                self.assertEqual(
                    self.client.mpub('foo', messages), ['response'])
                connection.mpub.assert_called_with('foo', messages)


if __name__ == '__main__':
    unittest.main()
